
import sys
import numpy as np
from MCEq.misc import theta_rad
import mceq_config as config

class EarthGeometry(object):
    r"""A model of the Earth's geometry, approximating it
       by a sphere. The figure below illustrates the meaning of the parameters.

    .. figure:: graphics/geometry.*
        :scale: 30 %
        :alt: picture of geometry


        Curved geometry as it is used in the code (not to scale!).

    Example:
      The plots below will be produced by executing the module::

          $ python geometry.py

    .. plot::

        import matplotlib.pyplot as plt
        import numpy as np
        from MCEq.geometry.geometry import *
        from MCEq.misc import theta_rad

        g = EarthGeometry()
        theta_list = np.linspace(0, 90, 500)
        h_vec = np.linspace(0, g.h_atm, 500)
        th_list_rad = theta_rad(theta_list)
        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        plt.plot(theta_list, g.l(th_list_rad) / 1e5,
                 lw=2)
        plt.xlabel(r'zenith $\theta$ at detector')
        plt.ylabel(r'path length $l(\theta)$ in km')
        ax = plt.gca()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')

        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        plt.plot(theta_list,
            np.arccos(g.cos_th_star(th_list_rad)) / np.pi * 180.,
            lw=2)
        plt.xlabel(r'zenith $\theta$ at detector')
        plt.ylabel(r'$\theta^*$ at top of the atm.')
        plt.ylim([0, 90])
        ax = plt.gca()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')

        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        plt.plot(h_vec / 1e5, g.delta_l(h_vec, theta_rad(85.)) / 1e5,
                 lw=2)
        plt.ylabel(r'Path length $\Delta l(h)$ in km')
        plt.xlabel(r'atm. height $h_{atm}$ in km')
        ax = plt.gca()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')

        fig = plt.figure(figsize=(5, 4))
        fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
        for theta in [30., 60., 70., 80., 85., 90.]:
            theta_path = theta_rad(theta)
            delta_l_vec = np.linspace(0, g.l(theta_path), 1000)
            plt.plot(delta_l_vec / 1e5, g.h(delta_l_vec, theta_path) / 1e5,
                     label=r'${0}^o$'.format(theta), lw=2)
        plt.legend()
        plt.xlabel(r'path length $\Delta l$ [km]')
        plt.ylabel(r'atm. height $h_{atm}(\Delta l, \theta)$ [km]')
        ax = plt.gca()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.xaxis.set_ticks_position('bottom')
        ax.yaxis.set_ticks_position('left')
        plt.show()

    Attributes:
      h_obs (float): observation level height [cm]
      h_atm (float): top of the atmosphere [cm]
      r_E (float): radius Earth [cm]
      r_top (float): radius at top of the atmosphere  [cm]
      r_obs (float): radius at observation level [cm]

    """
    def __init__(self):

        self.h_obs = config.h_obs * 1e2  # cm
        self.h_atm = config.h_atm * 1e2  # cm
        self.r_E = config.r_E * 1e2  # cm
        self.r_top = self.r_E + self.h_atm
        self.r_obs = self.r_E + self.h_obs

    def _A_1(self, theta):
        r"""Segment length :math:`A1(\theta)` in cm.
        """
        return self.r_obs * np.cos(theta)

    def _A_2(self, theta):
        r"""Segment length :math:`A2(\theta)` in cm.
        """
        return self.r_obs * np.sin(theta)

    def l(self, theta):
        r"""Returns path length in [cm] for given zenith
        angle :math:`\theta` [rad].
        """
        return (np.sqrt(self.r_top**2 - self._A_2(theta)**2) -
                self._A_1(theta))

    def cos_th_star(self, theta):
        r"""Returns the zenith angle at atmospheric boarder
        :math:`\cos(\theta^*)` in [rad] as a function of zenith at detector.
        """
        return (self._A_1(theta) + self.l(theta)) / self.r_top

    def h(self, dl, theta):
        r"""Height above surface at distance :math:`dl` counted from the beginning
        of path :math:`l(\theta)` in cm.
        """
        return np.sqrt(
            self._A_2(theta)**2 +
            (self._A_1(theta) + self.l(theta) - dl)**2) - self.r_E

    def delta_l(self, h, theta):
        r"""Distance :math:`dl` covered along path :math:`l(\theta)`
        as a function of current height. Inverse to :func:`h`.
        """
        return (self._A_1(theta) + self.l(theta) -
                np.sqrt((h + self.r_E)**2 - self._A_2(theta)**2))


def chirkin_cos_theta_star(costheta):
    r""":math:`\cos(\theta^*)` parameterization.

    This function returns the equivalent zenith angle for
    for very inclined showers. It is based on a CORSIKA study by
    `D. Chirkin, hep-ph/0407078v1, 2004
    <http://arxiv.org/abs/hep-ph/0407078v1>`_.

    Args:
        costheta (float): :math:`\cos(\theta)` in [rad]

    Returns:
        float: :math:`\cos(\theta*)` in [rad]
    """

    p1 = 0.102573
    p2 = -0.068287
    p3 = 0.958633
    p4 = 0.0407253
    p5 = 0.817285
    x = costheta
    return np.sqrt(
        (x**2 + p1**2 + p2 * x**p3 + p4 * x**p5) / (1 + p1**2 + p2 + p4))


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    earth = EarthGeometry()

    theta_list = np.linspace(0, 90, 500)
    h_vec = np.linspace(0, earth.h_atm, 500)
    th_list_rad = theta_rad(theta_list)
    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    plt.plot(theta_list, earth.l(th_list_rad) / 1e5, lw=2)
    plt.xlabel(r'zenith $\theta$ at detector')
    plt.ylabel(r'path length $l(\theta)$ in km')
    ax = plt.gca()
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    plt.plot(theta_list,
             np.arccos(earth.cos_th_star(th_list_rad)) / np.pi * 180.,
             lw=2)
    plt.xlabel(r'zenith $\theta$ at detector')
    plt.ylabel(r'$\theta^*$ at top of the atm.')
    plt.ylim([0, 90])
    ax = plt.gca()
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    plt.plot(h_vec / 1e5, earth.delta_l(h_vec, theta_rad(85.)) / 1e5, lw=2)
    plt.ylabel(r'Path length $\Delta l(h)$ in km')
    plt.xlabel(r'atm. height $h_{atm}$ in km')
    ax = plt.gca()
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')

    fig = plt.figure(figsize=(5, 4))
    fig.set_tight_layout(dict(rect=[0.00, 0.00, 1, 1]))
    for theta in [30., 60., 70., 80., 85., 90.]:
        theta_path = theta_rad(theta)
        delta_l_vec = np.linspace(0, earth.l(theta_path), 1000)
        plt.plot(delta_l_vec / 1e5,
                 earth.h(delta_l_vec, theta_path) / 1e5,
                 label=r'${0}^o$'.format(theta),
                 lw=2)
    plt.legend()
    plt.xlabel(r'path length $\Delta l$ [km]')
    plt.ylabel(r'atm. height $h_{atm}(\Delta l, \theta)$ [km]')
    ax = plt.gca()
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.xaxis.set_ticks_position('bottom')
    ax.yaxis.set_ticks_position('left')
    plt.show()




from abc import ABCMeta, abstractmethod
from six import with_metaclass
from os.path import join
import numpy as np
from MCEq.misc import theta_rad
from MCEq.misc import info

import mceq_config as config


class EarthsAtmosphere(with_metaclass(ABCMeta)):
    """
    Abstract class containing common methods on atmosphere.
    You have to inherit from this class and implement the virtual method
    :func:`get_density`.

    Note:
      Do not instantiate this class directly.

    Attributes:
      thrad (float): current zenith angle :math:`\\theta` in radiants
      theta_deg (float): current zenith angle :math:`\\theta` in degrees
      max_X (float): Slant depth at the surface according to the geometry
                     defined in the :mod:`MCEq.geometry`
      geometry (object): Can be a custom instance of EarthGeometry

    """

    def __init__(self, *args, **kwargs):
        from MCEq.geometry.geometry import EarthGeometry
        self.geom = kwargs.pop('geometry', EarthGeometry())
        self.thrad = None
        self.theta_deg = None
        self._max_den = config.max_density
        self.max_theta = 90.
        self.location = None
        self.season = None

    @abstractmethod
    def get_density(self, h_cm):
        """Abstract method which implementation  should return the density in g/cm**3.

        Args:
           h_cm (float):  height in cm

        Returns:
           float: density in g/cm**3

        Raises:
            NotImplementedError:
        """
        raise NotImplementedError("Base class called.")

    def calculate_density_spline(self, n_steps=2000):
        """Calculates and stores a spline of :math:`\\rho(X)`.

        Args:
          n_steps (int, optional): number of :math:`X` values
                                   to use for interpolation

        Raises:
            Exception: if :func:`set_theta` was not called before.
        """
        from scipy.integrate import cumtrapz
        from time import time
        from scipy.interpolate import UnivariateSpline

        if self.theta_deg is None:
            raise Exception('zenith angle not set')
        else:
            info(
                5, 'Calculating spline of rho(X) for zenith {0:4.1f} degrees.'.
                format(self.theta_deg))

        thrad = self.thrad
        path_length = self.geom.l(thrad)
        vec_rho_l = np.vectorize(
            lambda delta_l: self.get_density(self.geom.h(delta_l, thrad)))
        dl_vec = np.linspace(0, path_length, n_steps)

        now = time()

        # Calculate integral for each depth point
        X_int = cumtrapz(vec_rho_l(dl_vec), dl_vec)  #
        dl_vec = dl_vec[1:]

        info(5, '.. took {0:1.2f}s'.format(time() - now))

        # Save depth value at h_obs
        self._max_X = X_int[-1]
        self._max_den = self.get_density(self.geom.h(0, thrad))

        # Interpolate with bi-splines without smoothing
        h_intp = [self.geom.h(dl, thrad) for dl in reversed(dl_vec[1:])]
        X_intp = [X for X in reversed(X_int[1:])]

        self._s_h2X = UnivariateSpline(h_intp, np.log(X_intp), k=2, s=0.0)
        self._s_X2rho = UnivariateSpline(X_int, vec_rho_l(dl_vec), k=2, s=0.0)
        self._s_lX2h = UnivariateSpline(np.log(X_intp)[::-1],
                                       h_intp[::-1],
                                       k=2,
                                       s=0.0)

    @property
    def max_X(self):
        """Depth at altitude 0."""
        if not hasattr(self, '_max_X'):
            self.set_theta(0)
        return self._max_X

    @property
    def max_den(self):
        """Density at altitude 0."""
        if not hasattr(self, '_max_den'):
            self.set_theta(0)
        return self._max_den

    @property
    def s_h2X(self):
        """Spline for conversion from altitude to depth."""
        if not hasattr(self, '_s_h2X'):
            self.set_theta(0)
        return self._s_h2X
    @property
    def s_X2rho(self):
        """Spline for conversion from depth to density."""
        if not hasattr(self, '_s_X2rho'):
            self.set_theta(0)
        return self._s_X2rho
    @property
    def s_lX2h(self):
        """Spline for conversion from depth to altitude."""
        if not hasattr(self, '_s_lX2h'):
            self.set_theta(0)
        return self._s_lX2h

    def set_theta(self, theta_deg, force_spline_calc=False):
        """Configures geometry and initiates spline calculation for
        :math:`\\rho(X)`.

        If the option 'use_atm_cache' is enabled in the config, the
        function will check, if a corresponding spline is available
        in the cache and use it. Otherwise it will call
        :func:`calculate_density_spline`,  make the function
        :func:`r_X2rho` available to the core code and store the spline
        in the cache.

        Args:
          theta_deg (float): zenith angle :math:`\\theta` at detector
          force_spline_calc (bool): forces (re-)calculation of the
                                    spline for each call
        """
        if theta_deg < 0. or theta_deg > self.max_theta:
            raise Exception('Zenith angle not in allowed range.')

        self.thrad = theta_rad(theta_deg)
        self.theta_deg = theta_deg
        self.calculate_density_spline()

    def r_X2rho(self, X):
        """Returns the inverse density :math:`\\frac{1}{\\rho}(X)`.

        The spline `s_X2rho` is used, which was calculated or retrieved
        from cache during the :func:`set_theta` call.

        Args:
           X (float):  slant depth in g/cm**2

        Returns:
           float: :math:`1/\\rho` in cm**3/g

        """
        return 1. / self.s_X2rho(X)

    def h2X(self, h):
        """Returns the depth along path as function of height above
        surface.

        The spline `s_X2rho` is used, which was calculated or retrieved
        from cache during the :func:`set_theta` call.

        Args:
           h (float):  vertical height above surface in cm

        Returns:
           float: X  slant depth in g/cm**2

        """
        return np.exp(self.s_h2X(h))

    def X2h(self, X):
        """Returns the height above surface as a function of slant depth
        for currently selected zenith angle.

        The spline `s_lX2h` is used, which was calculated or retrieved
        from cache during the :func:`set_theta` call.

        Args:
           X (float):  slant depth in g/cm**2

        Returns:
           float h:  height above surface in cm

        """
        return self.s_lX2h(np.log(X))

    def X2rho(self, X):
        """Returns the density :math:`\\rho(X)`.

        The spline `s_X2rho` is used, which was calculated or retrieved
        from cache during the :func:`set_theta` call.

        Args:
           X (float):  slant depth in g/cm**2

        Returns:
           float: :math:`\\rho` in cm**3/g

        """
        return self.s_X2rho(X)

    def moliere_air(self, h_cm):
        """Returns the Moliere unit of air for US standard atmosphere. """

        return 9.3 / (self.get_density(h_cm) * 100.)

    def nref_rel_air(self, h_cm):
        """Returns the refractive index - 1 in air (density parametrization
        as in CORSIKA).
        """

        return 0.000283 * self.get_density(h_cm) / self.get_density(0)

    def gamma_cherenkov_air(self, h_cm):
        """Returns the Lorentz factor gamma of Cherenkov threshold in air (MeV).
        """

        nrel = self.nref_rel_air(h_cm)
        return (1. + nrel) / np.sqrt(2. * nrel + nrel**2)

    def theta_cherenkov_air(self, h_cm):
        """Returns the Cherenkov angle in air (degrees).
        """

        return np.arccos(1. / (1. + self.nref_rel_air(h_cm))) * 180. / np.pi


class CorsikaAtmosphere(EarthsAtmosphere):
    """Class, holding the parameters of a Linsley type parameterization
    similar to the Air-Shower Monte Carlo
    `CORSIKA <https://web.ikp.kit.edu/corsika/>`_.

    The parameters pre-defined parameters are taken from the CORSIKA
    manual. If new sets of parameters are added to :func:`init_parameters`,
    the array _thickl can be calculated using :func:`calc_thickl` .

    Attributes:
      _atm_param (numpy.array): (5x5) Stores 5 atmospheric parameters
                                _aatm, _batm, _catm, _thickl, _hlay
                                for each of the 5 layers

    Args:
      location (str): see :func:`init_parameters`
      season (str,optional): see :func:`init_parameters`
    """
    _atm_param = None

    def __init__(self, location, season=None):
        cka_atmospheres = [
            ("USStd", None),
            ("BK_USStd", None),
            ("Karlsruhe", None),
            ("ANTARES/KM3NeT-ORCA", 'Summer'),
            ("ANTARES/KM3NeT-ORCA", 'Winter'),
            ("KM3NeT-ARCA", 'Summer'),
            ("KM3NeT-ARCA", 'Winter'),
            ("KM3NeT",None),
            ('SouthPole','December'),
            ('PL_SouthPole','January'),
            ('PL_SouthPole','August'),
        ]
        assert (location, season) in cka_atmospheres, \
            '{0}/{1} not available for CorsikaAtmsophere'.format(
                location, season
            )
        self.init_parameters(location, season)
        import MCEq.geometry.corsikaatm.corsikaatm as corsika_acc
        self.corsika_acc = corsika_acc
        EarthsAtmosphere.__init__(self)

    def init_parameters(self, location, season):
        """Initializes :attr:`_atm_param`. Parameters from ANTARES/KM3NET
        are based on the work of T. Heid
        (`see this issue <https://github.com/afedynitch/MCEq/issues/12>`_)

        +---------------------+-------------------+------------------------------+
        | location            | CORSIKA Table     | Description/season           |
        +=====================+===================+==============================+
        | "USStd"             |         23        |  US Standard atmosphere      |
        +---------------------+-------------------+------------------------------+
        | "BK_USStd"          |         37        |  Bianca Keilhauer's USStd    |
        +---------------------+-------------------+------------------------------+
        | "Karlsruhe"         |         24        |  AT115 / Karlsruhe           |
        +---------------------+-------------------+------------------------------+
        | "SouthPole"         |      26 and 28    |  MSIS-90-E for Dec and June  |
        +---------------------+-------------------+------------------------------+
        |"PL_SouthPole"       |      29 and 30    |  P. Lipari's  Jan and Aug    |
        +---------------------+-------------------+------------------------------+
        |"ANTARES/KM3NeT-ORCA"|    NA             |  PhD T. Heid                 |
        +---------------------+-------------------+------------------------------+
        | "KM3NeT-ARCA"       |    NA             |  PhD T. Heid                 |
        +---------------------+-------------------+------------------------------+


        Args:
          location (str): see table
          season (str, optional): choice of season for supported locations

        Raises:
          Exception: if parameter set not available
        """
        _aatm, _batm, _catm, _thickl, _hlay = None, None, None, None, None
        if location == "USStd":
            _aatm = np.array([-186.5562, -94.919, 0.61289, 0.0, 0.01128292])
            _batm = np.array([1222.6562, 1144.9069, 1305.5948, 540.1778, 1.0])
            _catm = np.array([994186.38, 878153.55, 636143.04, 772170., 1.0e9])
            _thickl = np.array(
                [1036.102549, 631.100309, 271.700230, 3.039494, 0.001280])
            _hlay = np.array([0, 4.0e5, 1.0e6, 4.0e6, 1.0e7])
        elif location == "BK_USStd":
            _aatm = np.array([
                -149.801663, -57.932486, 0.63631894, 4.3545369e-4, 0.01128292
            ])
            _batm = np.array([1183.6071, 1143.0425, 1322.9748, 655.69307, 1.0])
            _catm = np.array(
                [954248.34, 800005.34, 629568.93, 737521.77, 1.0e9])
            _thickl = np.array(
                [1033.804941, 418.557770, 216.981635, 4.344861, 0.001280])
            _hlay = np.array([0.0, 7.0e5, 1.14e6, 3.7e6, 1.0e7])
        elif location == "Karlsruhe":
            _aatm = np.array(
                [-118.1277, -154.258, 0.4191499, 5.4094056e-4, 0.01128292])
            _batm = np.array([1173.9861, 1205.7625, 1386.7807, 555.8935, 1.0])
            _catm = np.array([919546., 963267.92, 614315., 739059.6, 1.0e9])
            _thickl = np.array(
                [1055.858707, 641.755364, 272.720974, 2.480633, 0.001280])
            _hlay = np.array([0.0, 4.0e5, 1.0e6, 4.0e6, 1.0e7])
        elif location == "KM3NeT": # averaged over detector and season
            _aatm = np.array([-141.31449999999998, -8.256029999999999, 0.6132505, -0.025998975, 0.4024275])
            _batm = np.array([1153.0349999999999, 1263.3325, 1257.0724999999998, 404.85974999999996, 1.0])
            _catm = np.array([967990.75, 668591.75, 636790.0, 814070.75, 21426175.0])
            _thickl = np.array([1011.8521512499999, 275.84507575000003, 51.0230705, 2.983134, 0.21927724999999998])
            _hlay = np.array([0.0, 993750.0, 2081250.0, 4150000.0, 6877500.0])
        elif location == "ANTARES/KM3NeT-ORCA":
            if season == 'Summer':
                _aatm = np.array([-158.85, -5.38682, 0.889893, -0.0286665, 0.50035])
                _batm = np.array([1145.62, 1176.79, 1248.92, 415.543, 1.0])
                _catm = np.array([998469.0, 677398.0, 636790.0, 823489.0, 16090500.0])
                _thickl = np.array([986.951713, 306.4668, 40.546793, 4.288721, 0.277182])
                _hlay = np.array([0, 9.0e5, 22.0e5, 38.0e5, 68.2e5])
            elif season == 'Winter':
                _aatm = np.array([-132.16, -2.4787, 0.298031, -0.0220264, 0.348021])
                _batm = np.array([1120.45, 1203.97, 1163.28, 360.027, 1.0])
                _catm = np.array([933697.0, 643957.0, 636790.0, 804486.0, 23109000.0])
                _thickl = np.array([988.431172, 273.033464, 37.185105, 1.162987, 0.192998])
                _hlay = np.array([0, 9.5e5, 22.0e5, 47.0e5, 68.2e5])
        elif location == "KM3NeT-ARCA":
            if season == 'Summer':
                _aatm = np.array([-157.857, -28.7524, 0.790275, -0.0286999, 0.481114])
                _batm = np.array([1190.44, 1171.0, 1344.78, 445.357, 1.0])
                _catm = np.array([1006100.0, 758614.0, 636790.0, 817384.0, 16886800.0])
                _thickl = np.array([1032.679434, 328.978681, 80.601135, 4.420745, 0.264112])
                _hlay = np.array([0, 9.0e5, 18.0e5, 38.0e5, 68.2e5])
            elif season == 'Winter':
                _aatm = np.array([-116.391, 3.5938, 0.474803, -0.0246031, 0.280225])
                _batm = np.array([1155.63, 1501.57, 1271.31, 398.512, 1.0])
                _catm = np.array([933697.0, 594398.0, 636790.0, 810924.0, 29618400.0])
                _thickl = np.array([1039.346286, 194.901358, 45.759249, 2.060083, 0.142817])
                _hlay = np.array([0, 12.25e5, 21.25e5, 43.0e5, 70.5e5])
        elif location == 'SouthPole':
            if season == 'December':
                _aatm = np.array(
                    [-128.601, -39.5548, 1.13088, -0.00264960, 0.00192534])
                _batm = np.array([1139.99, 1073.82, 1052.96, 492.503, 1.0])
                _catm = np.array(
                    [861913., 744955., 675928., 829627., 5.8587010e9])
                _thickl = np.array(
                    [1011.398804, 588.128367, 240.955360, 3.964546, 0.000218])
                _hlay = np.array([0.0, 4.0e5, 1.0e6, 4.0e6, 1.0e7])
            elif season == "June":
                _aatm = np.array(
                    [-163.331, -65.3713, 0.402903, -0.000479198, 0.00188667])
                _batm = np.array([1183.70, 1108.06, 1424.02, 207.595, 1.0])
                _catm = np.array(
                    [875221., 753213., 545846., 793043., 5.9787908e9])
                _thickl = np.array(
                    [1020.370363, 586.143464, 228.374393, 1.338258, 0.000214])
                _hlay = np.array([0.0, 4.0e5, 1.0e6, 4.0e6, 1.0e7])
            else:
                raise Exception('CorsikaAtmosphere(): Season "' + season +
                                '" not parameterized for location SouthPole.')
        elif location == 'PL_SouthPole':
            if season == 'January':
                _aatm = np.array(
                    [-113.139, -7930635, -54.3888, -0.0, 0.00421033])
                _batm = np.array([1133.10, 1101.20, 1085.00, 1098.00, 1.0])
                _catm = np.array(
                    [861730., 826340., 790950., 682800., 2.6798156e9])
                _thickl = np.array([
                    1019.966898, 718.071682, 498.659703, 340.222344, 0.000478
                ])
                _hlay = np.array([0.0, 2.67e5, 5.33e5, 8.0e5, 1.0e7])
            elif season == "August":
                _aatm = np.array(
                    [-59.0293, -21.5794, -7.14839, 0.0, 0.000190175])
                _batm = np.array([1079.0, 1071.90, 1182.0, 1647.1, 1.0])
                _catm = np.array(
                    [764170., 699910., 635650., 551010., 59.329575e9])
                _thickl = np.array(
                    [1019.946057, 391.739652, 138.023515, 43.687992, 0.000022])
                _hlay = np.array([0.0, 6.67e5, 13.33e5, 2.0e6, 1.0e7])
            else:
                raise Exception('CorsikaAtmosphere(): Season "' + season +
                                '" not parameterized for location SouthPole.')
        else:
            raise Exception("CorsikaAtmosphere:init_parameters(): Location " +
                            str(location) + " not parameterized.")

        self._atm_param = np.array([_aatm, _batm, _catm, _thickl, _hlay])

        self.location, self.season = location, season
        # Clear cached theta value to force spline recalculation
        self.theta_deg = None

    def depth2height(self, x_v):
        """Converts column/vertical depth to height.

        Args:
          x_v (float): column depth :math:`X_v` in g/cm**2

        Returns:
          float: height in cm
        """
        _aatm, _batm, _catm, _thickl, _hlay = self._atm_param

        if x_v >= _thickl[1]:
            height = _catm[0] * np.log(_batm[0] / (x_v - _aatm[0]))
        elif x_v >= _thickl[2]:
            height = _catm[1] * np.log(_batm[1] / (x_v - _aatm[1]))
        elif x_v >= _thickl[3]:
            height = _catm[2] * np.log(_batm[2] / (x_v - _aatm[2]))
        elif x_v >= _thickl[4]:
            height = _catm[3] * np.log(_batm[3] / (x_v - _aatm[3]))
        else:
            height = (_aatm[4] - x_v) * _catm[4]

        return height

    def get_density(self, h_cm):
        """ Returns the density of air in g/cm**3.

        Uses the optimized module function :func:`corsika_get_density_jit`.

        Args:
          h_cm (float): height in cm

        Returns:
          float: density :math:`\\rho(h_{cm})` in g/cm**3
        """
        return self.corsika_acc.corsika_get_density(h_cm, *self._atm_param)
        # return corsika_get_density_jit(h_cm, self._atm_param)

    def get_mass_overburden(self, h_cm):
        """ Returns the mass overburden in atmosphere in g/cm**2.

        Uses the optimized module function :func:`corsika_get_m_overburden_jit`

        Args:
          h_cm (float): height in cm

        Returns:
          float: column depth :math:`T(h_{cm})` in g/cm**2
        """
        return self.corsika_acc.corsika_get_m_overburden(h_cm, *self._atm_param)
        # return corsika_get_m_overburden_jit(h_cm, self._atm_param)

    def rho_inv(self, X, cos_theta):
        """Returns reciprocal density in cm**3/g using planar approximation.

        This function uses the optimized function :func:`planar_rho_inv_jit`

        Args:
          h_cm (float): height in cm

        Returns:
          float: :math:`\\frac{1}{\\rho}(X,\\cos{\\theta})` cm**3/g
        """
        return self.corsika_acc.planar_rho_inv(X, cos_theta, *self._atm_param)
        # return planar_rho_inv_jit(X, cos_theta, self._atm_param)

    def calc_thickl(self):
        """Calculates thickness layers for :func:`depth2height`

        The analytical inversion of the CORSIKA parameterization
        relies on the knowledge about the depth :math:`X`, where
        trasitions between layers/exponentials occur.

        Example:
          Create a new set of parameters in :func:`init_parameters`
          inserting arbitrary values in the _thikl array::

          $ cor_atm = CorsikaAtmosphere(new_location, new_season)
          $ cor_atm.calc_thickl()

          Replace _thickl values with printout.

        """
        from scipy.integrate import quad
        thickl = []
        for h in self._atm_param[4]:
            thickl.append('{0:4.6f}'.format(
                quad(self.get_density, h, 112.8e5, epsrel=1e-4)[0]))
        info(5, '_thickl = np.array([' + ', '.join(thickl) + '])')
        return thickl



class IsothermalAtmosphere(EarthsAtmosphere):
    """Isothermal model of the atmosphere.

    This model is widely used in semi-analytical calculations. The isothermal
    approximation is valid in a certain range of altitudes and usually
    one adjust the parameters to match a more realistic density profile
    at altitudes between 10 - 30 km, where the high energy muon production
    rate peaks. Such parametrizations are given in the book "Cosmic Rays and
    Particle Physics", Gaisser, Engel and Resconi (2016). The default values
    are from M. Thunman, G. Ingelman, and P. Gondolo, Astropart. Physics 5,
    309 (1996).

    Args:
      location (str): no effect
      season (str): no effect
      hiso_km (float): isothermal scale height in km
      X0 (float): Ground level overburden
    """

    def __init__(self, location, season, hiso_km=6.3, X0=1300.):
        self.hiso_cm = hiso_km * 1e5
        self.X0 = X0
        self.location = location
        self.season = season

        EarthsAtmosphere.__init__(self)

    def get_density(self, h_cm):
        """ Returns the density of air in g/cm**3.

        Args:
          h_cm (float): height in cm

        Returns:
          float: density :math:`\\rho(h_{cm})` in g/cm**3
        """
        return self.X0 / self.hiso_cm * np.exp(-h_cm / self.hiso_cm)

    def get_mass_overburden(self, h_cm):
        """ Returns the mass overburden in atmosphere in g/cm**2.

        Args:
          h_cm (float): height in cm

        Returns:
          float: column depth :math:`T(h_{cm})` in g/cm**2
        """
        return self.X0 * np.exp(-h_cm / self.hiso_cm)

class MSIS00Atmosphere(EarthsAtmosphere):
    """Wrapper class for a python interface to the NRLMSISE-00 model.

    `NRLMSISE-00 <http://ccmc.gsfc.nasa.gov/modelweb/atmos/nrlmsise00.html>`_
    is an empirical model of the Earth's atmosphere. It is available as
    a FORTRAN 77 code or as a verson traslated into
    `C by Dominik Borodowski <http://www.brodo.de/english/pub/nrlmsise/>`_.
    Here a PYTHON wrapper has been used.

    Attributes:
      _msis : NRLMSISE-00 python wrapper object handler

    Args:
      location (str): see :func:`init_parameters`
      season (str,optional): see :func:`init_parameters`
    """

    def __init__(self,
                 location,
                 season=None,
                 doy=None,
                 use_loc_altitudes=False):
        from MCEq.geometry.nrlmsise00_mceq import cNRLMSISE00

        msis_atmospheres = [
            'SouthPole',
            'Karlsruhe',
            'Geneva',
            'Tokyo',
            'SanGrasso',
            'TelAviv',
            'KSC',
            'SoudanMine',
            'Tsukuba',
            'LynnLake',
            'PeaceRiver',
            'FtSumner'
        ]
        assert location in msis_atmospheres, \
            '{0} not available for MSIS00Atmosphere'.format(
                location
            )

        self._msis = cNRLMSISE00()

        self.init_parameters(location, season, doy, use_loc_altitudes)

        EarthsAtmosphere.__init__(self)

    def init_parameters(self, location, season, doy, use_loc_altitudes):
        """Sets location and season in :class:`NRLMSISE-00`.

        Translates location and season into day of year
        and geo coordinates.

        Args:
          location (str): Supported are "SouthPole" and "Karlsruhe"
          season (str): months of the year: January, February, etc.
          use_loc_altitudes (bool): If to use default altitudes from location
        """
        self._msis.set_location(location)
        if season is not None:
            self._msis.set_season(season)
        else:
            self._msis.set_doy(doy)
        self.location, self.season = location, season
        # Clear cached value to force spline recalculation
        self.theta_deg = None
        if use_loc_altitudes:
            info(0, 'Using loc altitude', self._msis.alt_surface, 'cm')
            self.geom.h_obs = self._msis.alt_surface

    def get_density(self, h_cm):
        """ Returns the density of air in g/cm**3.

        Wraps around ctypes calls to the NRLMSISE-00 C library.

        Args:
          h_cm (float): height in cm

        Returns:
          float: density :math:`\\rho(h_{cm})` in g/cm**3
        """
        return self._msis.get_density(h_cm)

    def set_location(self, location):
        """ Changes MSIS location by strings defined in _msis_wrapper.

        Args:
          location (str): location as defined in :class:`NRLMSISE-00.`

        """
        self._msis.set_location(location)

    def set_season(self, month):
        """ Changes MSIS location by month strings defined in _msis_wrapper.

        Args:
          location (str): month as defined in :class:`NRLMSISE-00.`

        """
        self._msis.set_season(month)

    def set_doy(self, day_of_year):
        """ Changes MSIS season by day of year.

        Args:
          day_of_year (int): 1. Jan.=0, 1.Feb=32

        """
        self._msis.set_doy(day_of_year)

    def get_temperature(self, h_cm):
        """ Returns the temperature of air in K.

        Wraps around ctypes calls to the NRLMSISE-00 C library.

        Args:
          h_cm (float): height in cm

        Returns:
          float: density :math:`T(h_{cm})` in K
        """
        return self._msis.get_temperature(h_cm)


class AIRSAtmosphere(EarthsAtmosphere):
    """Interpolation class for tabulated atmospheres.

    This class is intended to read preprocessed AIRS Satellite data.

    Args:
      location (str): see :func:`init_parameters`
      season (str,optional): see :func:`init_parameters`
    """

    def __init__(self, location, season, extrapolate=True, *args, **kwargs):
        if location != 'SouthPole':
            raise Exception(self.__class__.__name__ +
                            "(): Only South Pole location supported. " +
                            location)

        self.extrapolate = extrapolate

        self.month2doy = {
            'January': 1,
            'February': 32,
            'March': 60,
            'April': 91,
            'May': 121,
            'June': 152,
            'July': 182,
            'August': 213,
            'September': 244,
            'October': 274,
            'November': 305,
            'December': 335
        }

        self.season = season
        self.init_parameters(location, **kwargs)
        EarthsAtmosphere.__init__(self)

    def init_parameters(self, location, **kwargs):
        """Loads tables and prepares interpolation.

        Args:
          location (str): supported is only "SouthPole"
          doy (int): Day Of Year
        """
        # from time import strptime
        from matplotlib.dates import datestr2num, num2date
        from os import path

        def bytespdate2num(b):
            return datestr2num(b.decode('utf-8'))

        data_path = (join(
            path.expanduser('~'),
            'OneDrive/Dokumente/projects/atmospheric_variations/'))

        if 'table_path' in kwargs:
            data_path = kwargs['table_path']

        files = [('dens', 'airs_amsu_dens_180_daily.txt'),
                 ('temp', 'airs_amsu_temp_180_daily.txt'),
                 ('alti', 'airs_amsu_alti_180_daily.txt')]

        data_collection = {}

        # limit SouthPole pressure to <= 600
        min_press_idx = 4

        IC79_idx_1 = None
        IC79_idx_2 = None

        for d_key, fname in files:
            fname = data_path + 'tables/' + fname
            # tabf = open(fname).read()

            tab = np.loadtxt(fname,
                             converters={0: bytespdate2num},
                             usecols=[0] + list(range(2, 27)))
            # with open(fname, 'r') as f:
            #     comline = f.readline()
            # p_levels = [
            #     float(s.strip()) for s in comline.split(' ')[3:] if s != ''
            # ][min_press_idx:]
            dates = num2date(tab[:, 0])
            for di, date in enumerate(dates):
                if date.month == 6 and date.day == 1:
                    if date.year == 2010:
                        IC79_idx_1 = di
                    elif date.year == 2011:
                        IC79_idx_2 = di
            surf_val = tab[:, 1]
            cols = tab[:, min_press_idx + 2:]
            data_collection[d_key] = (dates, surf_val, cols)

        self.interp_tab_d = {}
        self.interp_tab_t = {}
        self.dates = {}
        dates = data_collection['alti'][0]

        msis = MSIS00Atmosphere(location, 'January')
        for didx, date in enumerate(dates):
            h_vec = np.array(data_collection['alti'][2][didx, :] * 1e2)
            d_vec = np.array(data_collection['dens'][2][didx, :])
            t_vec = np.array(data_collection['temp'][2][didx, :])

            if self.extrapolate:
                # Extrapolate using msis
                h_extra = np.linspace(h_vec[-1], self.geom.h_atm * 1e2, 250)
                msis._msis.set_doy(self._get_y_doy(date)[1] - 1)
                msis_extra_d = np.array([msis.get_density(h) for h in h_extra])
                msis_extra_t = np.array(
                    [msis.get_temperature(h) for h in h_extra])

                # Interpolate last few altitude bins
                ninterp = 5

                for ni in range(ninterp):
                    cl = (1 - np.exp(-ninterp + ni + 1))
                    ch = (1 - np.exp(-ni))
                    norm = 1. / (cl + ch)
                    d_vec[-ni -
                          1] = (d_vec[-ni - 1] * cl * norm +
                                msis.get_density(h_vec[-ni - 1]) * ch * norm)
                    t_vec[-ni - 1] = (
                        t_vec[-ni - 1] * cl * norm +
                        msis.get_temperature(h_vec[-ni - 1]) * ch * norm)

                # Merge the two datasets
                h_vec = np.hstack([h_vec[:-1], h_extra])
                d_vec = np.hstack([d_vec[:-1], msis_extra_d])
                t_vec = np.hstack([t_vec[:-1], msis_extra_t])

            self.interp_tab_d[self._get_y_doy(date)] = (h_vec, d_vec)
            self.interp_tab_t[self._get_y_doy(date)] = (h_vec, t_vec)

            self.dates[self._get_y_doy(date)] = date

        self.IC79_start = self._get_y_doy(dates[IC79_idx_1])
        self.IC79_end = self._get_y_doy(dates[IC79_idx_2])
        self.IC79_days = (dates[IC79_idx_2] - dates[IC79_idx_1]).days
        self.location = location
        if self.season is None:
            self.set_IC79_day(0)
        else:
            self.set_season(self.season)
        # Clear cached value to force spline recalculation
        self.theta_deg = None

    def set_date(self, year, doy):
        self.h, self.dens = self.interp_tab_d[(year, doy)]
        _, self.temp = self.interp_tab_t[(year, doy)]
        self.date = self.dates[(year, doy)]
        # Compatibility with caching
        self.season = self.date

    def _set_doy(self, doy, year=2010):
        self.h, self.dens = self.interp_tab_d[(year, doy)]
        _, self.temp = self.interp_tab_t[(year, doy)]
        self.date = self.dates[(year, doy)]

    def set_season(self, month):
        self.season = month
        self._set_doy(self.month2doy[month])
        self.season = month

    def set_IC79_day(self, IC79_day):
        import datetime
        if IC79_day > self.IC79_days:
            raise Exception(self.__class__.__name__ +
                            "::set_IC79_day(): IC79_day above range.")
        target_day = self._get_y_doy(self.dates[self.IC79_start] +
                                     datetime.timedelta(days=IC79_day))
        info(2, 'setting IC79_day', IC79_day)
        self.h, self.dens = self.interp_tab_d[target_day]
        _, self.temp = self.interp_tab_t[target_day]
        self.date = self.dates[target_day]
        # Compatibility with caching
        self.season = self.date

    def _get_y_doy(self, date):
        return date.timetuple().tm_year, date.timetuple().tm_yday

    def get_density(self, h_cm):
        """ Returns the density of air in g/cm**3.

        Interpolates table at requested value for previously set
        year and day of year (doy).

        Args:
          h_cm (float): height in cm

        Returns:
          float: density :math:`\\rho(h_{cm})` in g/cm**3
        """
        ret = np.exp(np.interp(h_cm, self.h, np.log(self.dens)))
        try:
            ret[h_cm > self.h[-1]] = np.nan
        except TypeError:
            if h_cm > self.h[-1]:
                return np.nan
        return ret

    def get_temperature(self, h_cm):
        """ Returns the temperature in K.

        Interpolates table at requested value for previously set
        year and day of year (doy).

        Args:
          h_cm (float): height in cm

        Returns:
          float: temperature :math:`T(h_{cm})` in K
        """
        ret = np.exp(np.interp(h_cm, self.h, np.log(self.temp)))
        try:
            ret[h_cm > self.h[-1]] = np.nan
        except TypeError:
            if h_cm > self.h[-1]:
                return np.nan
        return ret


class MSIS00IceCubeCentered(MSIS00Atmosphere):
    """Extension of :class:`MSIS00Atmosphere` which couples the latitude
    setting with the zenith angle of the detector.

    Args:
      location (str): see :func:`init_parameters`
      season (str,optional): see :func:`init_parameters`
    """

    def __init__(self, location, season):
        if location != 'SouthPole':
            info(2, 'location forced to the South Pole')
            location = 'SouthPole'

        MSIS00Atmosphere.__init__(self, location, season)

        # Allow for upgoing zenith angles
        self.max_theta = 180.

    def latitude(self, det_zenith_deg):
        """ Returns the geographic latitude of the shower impact point.

        Assumes a spherical earth. The detector is 1948m under the
        surface.

        Credits: geometry fomulae by Jakob van Santen, DESY Zeuthen.

        Args:
          det_zenith_deg (float): zenith angle at detector in degrees

        Returns:
          float: latitude of the impact point in degrees
        """
        r = self.geom.r_E
        d = 1948  # m

        theta_rad = det_zenith_deg / 180. * np.pi

        x = (np.sqrt(2. * r * d + ((r - d) * np.cos(theta_rad))**2 - d**2) -
             (r - d) * np.cos(theta_rad))

        return -90. + np.arctan2(x * np.sin(theta_rad),
                                 r - d + x * np.cos(theta_rad)) / np.pi * 180.

    def set_theta(self, theta_deg, force_spline_calc=True):

        self._msis.set_location_coord(longitude=0.,
                                      latitude=self.latitude(theta_deg))
        info(
            1, 'latitude = {0:5.2f} for zenith angle = {1:5.2f}'.format(
                self.latitude(theta_deg), theta_deg))
        if theta_deg > 90.:
            info(
                1, 'theta = {0:5.2f} below horizon. using theta = {1:5.2f}'.
                format(theta_deg, 180. - theta_deg))
            theta_deg = 180. - theta_deg
        MSIS00Atmosphere.set_theta(self,
                                   theta_deg,
                                   force_spline_calc=force_spline_calc)


class GeneralizedTarget(object):
    """This class provides a way to run MCEq on piece-wise constant
    one-dimenional density profiles.

    The default values for the average density are taken from
    config file variables `len_target`, `env_density` and `env_name`.
    The density profile has to be built by calling subsequently
    :func:`add_material`. The current composition of the target
    can be checked with :func:`draw_materials` or :func:`print_table`.

    Note:
      If the target is not air or hydrogen, the result is approximate,
      since seconray particle yields are provided for nucleon-air or
      proton-proton collisions. Depending on this choice one has to
      adjust the nuclear mass in :mod:`mceq_config`.

    Args:
      len_target (float): total length of the target in meters
      env_density (float): density of the default material in g/cm**3
      env_name (str): title for this environment
    """

    def __init__(
            self,
            len_target=config.len_target * 1e2,  # cm
            env_density=config.env_density,  # g/cm3
            env_name=config.env_name):

        self.len_target = len_target
        self.env_density = env_density
        self.env_name = env_name
        self.reset()

    @property
    def max_den(self):
        return self._max_den

    def reset(self):
        """Resets material list to defaults."""
        self.mat_list = [[
            0., self.len_target, self.env_density, self.env_name
        ]]
        self._update_variables()

    def _update_variables(self):
        """Updates internal variables. Not needed to call by user."""

        self.start_bounds, self.end_bounds, \
            self.densities = list(zip(*self.mat_list))[:-1]
        self.densities = np.array(self.densities)
        self.start_bounds = np.array(self.start_bounds)
        self.end_bounds = np.array(self.end_bounds)
        self._max_den = np.max(self.densities)
        self._integrate()

    def set_length(self, new_length_cm):
        """Updates the total length of the target.

        Usually the length is set
        """
        if new_length_cm < self.mat_list[-1][0]:
            raise Exception(
                "GeneralizedTarget::set_length(): " +
                "can not set length below lower boundary of last " +
                "material.")
        self.len_target = new_length_cm
        self.mat_list[-1][1] = new_length_cm
        self._update_variables()

    def add_material(self, start_position_cm, density, name):
        """Adds one additional material to a composite target.

        Args:
           start_position_cm (float):  position where the material starts
                                       counted from target origin l|X = 0 in cm
           density (float):  density of material in g/cm**3
           name (str):  any user defined name

        Raises:
            Exception: If requested start_position_cm is not properly defined.
        """

        if start_position_cm < 0. or start_position_cm > self.len_target:
            raise Exception("GeneralizedTarget::add_material(): " +
                            "distance exceeds target dimensions.")
        elif (start_position_cm == self.mat_list[-1][0]
              and self.mat_list[-1][-1] == self.env_name):
            self.mat_list[-1] = [
                start_position_cm, self.len_target, density, name
            ]

        elif start_position_cm <= self.mat_list[-1][0]:
            raise Exception("GeneralizedTarget::add_material(): " +
                            "start_position_cm is ahead of previous material.")

        else:
            self.mat_list[-1][1] = start_position_cm
            self.mat_list.append(
                [start_position_cm, self.len_target, density, name])

        info(2,
             ("{0}::add_material(): Material '{1}' added. " +
              "location on path {2} to {3} m").format(self.__class__.__name__,
                                                      name,
                                                      self.mat_list[-1][0],
                                                      self.mat_list[-1][1]))

        self._update_variables()

    def set_theta(self, *args):
        """This method is not defined for the generalized target. The purpose
        is to catch usage errors.

        Raises:
            NotImplementedError: always
        """

        raise NotImplementedError('GeneralizedTarget::set_theta(): Method' +
                                  'not defined for this target class.')

    def _integrate(self):
        """Walks through material list and computes the depth along the
        position (path). Computes the spline for the position-depth relation
        and determines the maximum depth for the material selection.

        Method does not need to be called by the user, instead the class
        calls it when necessary.
        """

        from scipy.interpolate import UnivariateSpline
        self.density_depth = None
        self.knots = [0.]
        self.X_int = [0.]

        for start, end, density, _ in self.mat_list:
            self.knots.append(end)
            self.X_int.append(density * (end - start) + self.X_int[-1])

        self._s_X2h = UnivariateSpline(self.X_int, self.knots, k=1, s=0.)
        self._s_h2X = UnivariateSpline(self.knots, self.X_int, k=1, s=0.)
        self._max_X = self.X_int[-1]

    @property
    def s_X2h(self):
        """Spline for depth at distance."""
        if not hasattr(self, '_s_X2h'):
            self._integrate()
        return self._s_X2h

    @property
    def s_h2X(self):
        """Spline for distance at depth."""
        if not hasattr(self, '_s_h2X'):
            self._integrate()
        return self._s_h2X

    @property
    def max_X(self):
        """Maximal depth of target."""
        if not hasattr(self, '_max_X'):
            self._integrate()
        return self._max_X

    def get_density_X(self, X):
        """Returns the density in g/cm**3 as a function of depth X.

        Args:
           X (float):  depth in g/cm**2

        Returns:
           float: density in g/cm**3

        Raises:
            Exception: If requested depth exceeds target.
        """
        X = np.atleast_1d(X)
        # allow for some small constant extrapolation for odepack solvers
        if X[-1] > self.max_X and X[-1] < self.max_X * 1.003:
            X[-1] = self.max_X
        if np.min(X) < 0. or np.max(X) > self.max_X:
            # return self.get_density(self.s_X2h(self.max_X))
            info(0, 'Depth {0:4.3f} exceeds target dimensions {1:4.3f}'.format(
                np.max(X), self.max_X
            ))
            raise Exception('Invalid input')

        return self.get_density(self.s_X2h(X))

    def r_X2rho(self, X):
        """Returns the inverse density :math:`\\frac{1}{\\rho}(X)`.

        Args:
           X (float):  slant depth in g/cm**2

        Returns:
           float: :math:`1/\\rho` in cm**3/g

        """
        return 1. / self.get_density_X(X)

    def get_density(self, l_cm):
        """Returns the density in g/cm**3 as a function of position l in cm.

        Args:
           l (float):  position in target in cm

        Returns:
           float: density in g/cm**3

        Raises:
            Exception: If requested position exceeds target length.
        """
        l_cm = np.atleast_1d(l_cm)
        res = np.zeros_like(l_cm)

        if np.min(l_cm) < 0 or np.max(l_cm) > self.len_target:
            raise Exception("GeneralizedTarget::get_density(): " +
                            "requested position exceeds target legth.")
        for i, li in enumerate(l_cm):
            bi = 0
            while not (li >= self.start_bounds[bi]
                       and li <= self.end_bounds[bi]):
                bi += 1
            res[i] = self.densities[bi]
        return res

    def draw_materials(self, axes=None, logx=False):
        """Makes a plot of depth and density profile as a function
        of the target length. The list of materials is printed out, too.

        Args:
           axes (plt.axes, optional):  handle for matplotlib axes
        """
        import matplotlib.pyplot as plt

        if not axes:
            plt.figure(figsize=(5, 2.5))
            axes = plt.gca()
        ymax = np.max(self.X_int) * 1.01
        for _, mat in enumerate(self.mat_list):
            xstart = mat[0]
            xend = mat[1]
            alpha = 0.188 * mat[2] / max(self.densities) + 0.248
            if alpha > 1:
                alpha = 1.
            elif alpha < 0.:
                alpha = 0.
            axes.fill_between((xstart, xend), (ymax, ymax), (0., 0.),
                              label=mat[2],
                              facecolor='grey',
                              alpha=alpha)
            # axes.text(0.5e-2 * (xstart + xend), 0.5 * ymax, str(nm))

        axes.plot([xl for xl in self.knots], self.X_int, lw=1.7, color='r')

        if logx:
            axes.set_xscale('log', nonposx='clip')

        axes.set_ylim(0., ymax)
        axes.set_xlabel('distance in target (cm)')
        axes.set_ylabel(r'depth X (g/cm$^2)$')

        self.print_table(min_dbg_lev=2)

    def print_table(self, min_dbg_lev=0):
        """Prints table of materials to standard output.
        """

        templ = '{0:^3} | {1:15} | {2:^9.3g}  | {3:^9.3g} | {4:^8.5g}'
        info(
            min_dbg_lev,
            '********************* List of materials ***********************',
            no_caller=True)
        head = '{0:3} | {1:15} | {2:9} | {3:9} | {4:9}'.format(
            'no', 'name', 'start [cm]', 'end [cm]', 'density [g/cm**3]')
        info(min_dbg_lev, '-' * len(head), no_caller=True)
        info(min_dbg_lev, head, no_caller=True)
        info(min_dbg_lev, '-' * len(head), no_caller=True)
        for nm, mat in enumerate(self.mat_list):
            info(min_dbg_lev,
                 templ.format(nm, mat[3], mat[0], mat[1], mat[2]),
                 no_caller=True)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    plt.figure(figsize=(5, 4))
    plt.title('CORSIKA atmospheres')
    cka_atmospheres = [
        ("USStd", None),
        ("BK_USStd", None),
        ("Karlsruhe", None),
        ("ANTARES/KM3NeT-ORCA", 'Summer'),
        ("ANTARES/KM3NeT-ORCA", 'Winter'),
        ("KM3NeT-ARCA", 'Summer'),
        ("KM3NeT-ARCA", 'Winter'),
        ("KM3NeT", None),
        ('SouthPole','December'),
        ('PL_SouthPole','January'),
        ('PL_SouthPole','August'),
    ]
    cka_surf_100 = []
    for loc, season in cka_atmospheres:
        cka_obj = CorsikaAtmosphere(loc, season)
        cka_obj.set_theta(0.0)
        x_vec = np.linspace(0, cka_obj.max_X, 5000)
        plt.plot(x_vec,
                1 / cka_obj.r_X2rho(x_vec),
                lw=1.5,
                label='{0}/{1}'.format(loc, season) if season is not None
                    else '{0}'.format(loc))
        cka_surf_100.append((cka_obj.max_X, 1. / cka_obj.r_X2rho(100.)))
    print(cka_surf_100)
    plt.ylabel(r'Density $\rho$ (g/cm$^3$)')
    plt.xlabel(r'Depth (g/cm$^2$)')
    plt.legend(loc='upper left')
    plt.tight_layout()

    plt.figure(figsize=(5, 4))
    plt.title('NRLMSISE-00 atmospheres')
    msis_atmospheres = [
        ('SouthPole', "January"),
        ('Karlsruhe', "January"),
        ('Geneva', "January"),
        ('Tokyo', "January"),
        ('SanGrasso', "January"),
        ('TelAviv', "January"),
        ('KSC', "January"),
        ('SoudanMine', "January"),
        ('Tsukuba', "January"),
        ('LynnLake', "January"),
        ('PeaceRiver', "January"),
        ('FtSumner', "January")
    ]
    msis_surf_100 = []
    for loc, season in msis_atmospheres:
        msis_obj = MSIS00Atmosphere(loc, season)
        msis_obj.set_theta(0.0)
        x_vec = np.linspace(0, msis_obj.max_X, 5000)
        plt.plot(x_vec,
                1 / msis_obj.r_X2rho(x_vec),
                lw=1.5,
                label='{0}'.format(loc))
        msis_surf_100.append((msis_obj.max_X, 1. / msis_obj.r_X2rho(100.)))
    print(msis_surf_100)
    plt.ylabel(r'Density $\rho$ (g/cm$^3$)')
    plt.xlabel(r'Depth (g/cm$^2$)')
    plt.legend(loc='upper left')
    plt.tight_layout()
    plt.show()



from __future__ import print_function
import sys
import platform
import os.path as path
import warnings
base_path = path.dirname(path.abspath(__file__))

#: Debug flag for verbose printing, 0 silences MCEq entirely
debug_level = 1
#: Override debug prinput for functions listed here (just give the name,
#: "get_solution" for instance) Warning, this option slows down initialization
#: by a lot. Use only when needed.
override_debug_fcn = []
#: Override debug printout for debug levels < value for the functions above
override_max_level = 10
#: Print module name in debug output
print_module = False

# =================================================================
# Paths and library locations
# =================================================================

#: Directory where the data files for the calculation are stored
data_dir = path.join(base_path, 'MCEq', 'data')

#: File name of the MCEq database
mceq_db_fname = "mceq_db_lext_dpm191_v12.h5"

#: File name of the MCEq database
em_db_fname = "mceq_db_EM_Tsai-Max_Z7.31.h5"

# =================================================================
# Atmosphere and geometry settings
# =================================================================

#: The latest versions of MCEq work in kinetic energy not total energy
#: If you want the result to be compatible with the previous choose
#: 'total energy' else 'kinetic energy'
return_as = "kinetic energy"
#: Atmospheric model in the format: (model, (arguments))
density_model = ('CORSIKA', ('BK_USStd', None))
#: density_model = ('MSIS00_IC',('SouthPole','January'))
#: density_model = ('GeneralizedTarget', None)

#: Definition of prompt: default ctau < 0.123 cm (that of D0)
prompt_ctau = 0.123

#: Average mass of target (for interaction length calculations)
#: Change parameter only in combination with interaction model setting.
#: By default all particle production matrices are calculated for air targets
#: expect those for models with '_pp' suffix. These are valid for hydrogen targets.
#: <A> = 14.6568 for air as below (source https://en.wikipedia.org/wiki/Atmosphere_of_Earth)
A_target = sum([f[0]*f[1] for f in [(0.78084, 14), (0.20946, 16), (0.00934, 40)]])

#: parameters for EarthGeometry
r_E = 6391.e3  # Earth radius in m
h_obs = 0.  # observation level in m
h_atm = 112.8e3  # top of the atmosphere in m

#: Default parameters for GeneralizedTarget
#: Total length of the target [m]
len_target = 1000.
#: density of default material in g/cm^3
env_density = 0.001225
env_name = "air"
#: Approximate value for the maximum density expected. Needed for the
#: resonance approximation. Default value: air at the surface
max_density = 0.001225,
#: Material for ionization and radiation (=continuous) loss terms
#: Currently available choices: 'air', 'water', 'ice'
dedx_material = 'air'
# =================================================================
# Parameters of numerical integration
# =================================================================

#: Minimal energy for grid
#: The minimal energy (technically) is 1e-2 GeV. Currently you can run into
#: stability problems with the integrator with such low thresholds. Use with
#: care and check results for oscillations and feasibility.
e_min = .1

#: The maximal energy is 1e12 GeV, but not all interaction models run at such
#: high energies. If you are interested in lower energies, reduce this value
#: for inclusive calculations to max. energy of interest + 4-5 orders of
#: magnitude. For single primaries the maximal energy is directly limited by
#: this value. Smaller grids speed up the initialization and integration.
e_max = 1e11

#: Enable electromagnetic cascade with matrices from EmCA
enable_em = False

#: Selection of integrator (euler/odepack)
integrator = "euler"

#: euler kernel implementation (numpy/MKL/CUDA).
#: With serious nVidia GPUs CUDA a few times faster than MKL
#: autodetection of fastest kernel below
kernel_config = "numpy"

#: Select CUDA device ID if you have multiple GPUs
cuda_gpu_id = 0

#: CUDA Floating point precision (default 32-bit 'float')
cuda_fp_precision = 32

#: Number of MKL threads (for sparse matrix multiplication the performance
#: advantage from using more than a few threads is limited by memory bandwidth)
#: Irrelevant for GPU integrators, but can affect initialization speed if
#: numpy is linked to MKL. 
mkl_threads = 8

#: parameters for the odepack integrator. More details at
#: http://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.ode.html#scipy.integrate.ode
ode_params = {
    'name': 'lsoda',
    'method': 'bdf',
    'nsteps': 1000,
    # 'max_step': 10.0,
    'rtol': 0.01
}

# =========================================================================
# Advanced settings
# =========================================================================

#: The leading process is can be either "decays" or "interactions". This depends
#: on the target density and it is usually chosen automatically. For
#: advanced applications one can force "interactions" to be the dominant
#: process. Essentially this affects how the adaptive step size is computed.
#: There is also the choice of "auto" that takes both processes into account
leading_process = "auto"

#: Stability margin for the integrator. The default 0.95 means that step
#: sizes are chosen 5% away from the stability circle. Usually no need to
#: change, except you know what it does.
stability_margin = 0.95

#: Ratio of decay_length/interaction_length where particle interactions
#: are neglected and the resonance approximation is used
#: 0.5 ~ precision loss <+3% speed gain ~ factor 10
#: If smoothness and shape accuracy for prompt flux is crucial, use smaller
#: values around 0.1 or 0.05
hybrid_crossover = 0.5

#: Maximal integration step dX in g/cm2. No limit necessary in most cases,
#: use for debugging purposes when searching for stability issues.
dXmax = 10.

#: Enable default tracking particles, such as pi_numu, pr_mu+, etc.
#: If only total fluxes are of interest, disable this feature to gain
#: performance since the eqution system becomes smaller and sparser
enable_default_tracking = True

#: Muon energy loss according to Kokoulin et al.
enable_muon_energy_loss = True

#: enable EM ionization loss
enable_em_ion = False

#: Improve (explicit solver) stability by averaging the continous loss
#: operator
average_loss_operator = True

#: Step size (dX) for averaging
loss_step_for_average = 1e-1

#: Raise exception when requesting unknown particles from get_solution
excpt_on_missing_particle = False

#: When using modified particle production matrices use
#: isospin symmetries to determine the corresponding
#: modification for neutrons and K0L/K0S
use_isospin_sym = True

#: Helicity dependent muons decays from analytical expressions
muon_helicity_dependence = True

#: Assume nucleon, pion and kaon cross sections for interactions of
#: rare or exotic particles (mostly relevant for non-compact mode)
assume_nucleon_interactions_for_exotics = True

#: This is not used in the code as before, instead the low energy
#: extension is compiled into the HDF backend files.
low_energy_extension = {
    "he_le_transition": 80,  # GeV
    "nbins_interp": 3,
    "use_unknown_cs": True,
}

#: Advanced settings (some options might be obsolete/not working)
adv_set = {
    #: Disable particle production by all hadrons, except nucleons
    "disable_interactions_of_unstable": False,

    #: Disable particle production by charm *projectiles* (interactions)
    "disable_charm_pprod": False,

    #: Disable resonance/prompt contribution (this group of options
    #: is either obsolete or needs maintenance.)
    #: "disable_resonance_decay" : False,

    #: Allow only those particles to be projectiles (incl. anti-particles)
    #: Faster initialization,
    #: For inclusive lepton flux computations:
    #: precision loss ~ 1%, for SIBYLL2.3.X with charm 5% above 10^7 GeV
    #: Might be different for yields (set_single_primary_particle)
    #: For full precision or if in doubt, use []
    "allowed_projectiles": [],  # [2212, 2112, 211, 321, 130, 11, 22],

    #: Disable particle (production) 
    "disabled_particles": [], #20, 19, 18, 17, 97, 98, 99, 101, 102, 103

    #: Disable leptons coming from prompt hadron decays at the vertex
    "disable_direct_leptons": False,

    #: Difficult to explain parameter
    'disable_leading_mesons': False,

    #: Do not apply mixing to these particles
    "exclude_from_mixing": [13],

    #: Switch off decays. E.g., disable muon decay with [13,-13]
    "disable_decays": [],

    #: Force particles to be treated as resonance
    "force_resonance": [],

    #: Disable mixing between resonance approx. and full propagation
    "no_mixing": False
}

#: Particles for compact mode
standard_particles = [
    11, 12, 13, 14, 16, 211, 321, 2212, 2112, 3122, 411, 421, 431
]

#: Anti-particles
standard_particles += [-pid for pid in standard_particles]

#: unflavored particles
#: append 221, 223, 333, if eta, omega and phi needed directly
standard_particles += [22, 111, 130, 310]  #: , 221, 223, 333]

#: This construct provides access to the attributes as in previous
#: versions, using `from mceq_config import config`. The future versions
#: will access the module attributes directly.

#: Autodetect best solver
#: determine shared library extension and MKL path
pf = platform.platform()

if 'Linux' in pf:
    mkl_path = path.join(sys.prefix, 'lib', 'libmkl_rt.so')
elif 'Darwin' in pf:
    mkl_path = path.join(sys.prefix, 'lib', 'libmkl_rt.dylib')
else:
    # Windows case
    mkl_path = path.join(sys.prefix, 'Library', 'bin', 'mkl_rt.dll')

# mkl library handler
mkl = None

# Check if MKL library found
if path.isfile(mkl_path):
    has_mkl = True
else:
    has_mkl = False

# Look for cupy module
try:
    import cupy
    has_cuda = True
except ImportError:
    has_cuda = False

# CUDA is usually fastest, then MKL. Fallback to numpy.
if has_cuda:
    kernel_config = 'CUDA'
elif has_mkl:
    kernel_config = 'MKL'
else:
    kernel_config = 'numpy'
if debug_level >= 2:
    print('Auto-detected {0} solver.'.format(kernel_config))

def set_mkl_threads(nthreads):
    global mkl_threads, mkl
    from ctypes import cdll, c_int, byref
    mkl = cdll.LoadLibrary(mkl_path)
    # Set number of threads
    mkl_threads = nthreads
    mkl.mkl_set_num_threads(byref(c_int(nthreads)))
    if debug_level >= 5:
        print('MKL threads limited to {0}'.format(nthreads))

if has_mkl:
    set_mkl_threads(mkl_threads)



# Compatibility layer for dictionary access to config attributes
# This is deprecated and will be removed in future


class MCEqConfigCompatibility(dict):
    """This class provides access to the attributes of the module as a
    dictionary, as it was in the previous versions of MCEq

    This method is deprecated and will be removed in future.
    """

    def __init__(self, namespace):
        self.__dict__.update(namespace)
        if debug_level > 1:
            warn_str = ("Config dictionary is deprecated. " +
                        "Use config.variable instead of config['variable']")
            warnings.warn(warn_str, FutureWarning)

    def __setitem__(self, key, value):
        key = key.lower()
        if key not in self.__dict__:
            raise Exception('Unknown config key', key)
        return super(MCEqConfigCompatibility, self).__setitem__(key, value)


config = MCEqConfigCompatibility(globals())


def _download_file(url, outfile):
    """Downloads the MCEq database from github"""

    from tqdm import tqdm
    import requests
    import math

    # Streaming, so we can iterate over the response.
    r = requests.get(url, stream=True)

    # Total size in bytes.
    total_size = int(r.headers.get('content-length', 0))
    block_size = 1024 * 1024
    wrote = 0
    with open(outfile, 'wb') as f:
        for data in tqdm(r.iter_content(block_size), total=math.ceil(total_size // block_size),
                         unit='MB', unit_scale=True):
            wrote = wrote + len(data)
            f.write(data)
    if total_size != 0 and wrote != total_size:
        raise Exception("ERROR, something went wrong")


# Download database file from github
base_url = 'https://github.com/afedynitch/MCEq/releases/download/'
release_tag = 'builds_on_azure/'
url = base_url + release_tag + mceq_db_fname
if not path.isfile(path.join(data_dir, mceq_db_fname)):
    print('Downloading for mceq database file {0}.'.format(mceq_db_fname))
    if debug_level >= 2:
        print(url)
    _download_file(url, path.join(data_dir, mceq_db_fname))

if path.isfile(path.join(data_dir, 'mceq_db_lext_dpm191.h5')):
    import os
    print('Removing previous database {0}.'.format('mceq_db_lext_dpm191.h5'))
    os.unlink(path.join(data_dir, 'mceq_db_lext_dpm191.h5'))



from __future__ import print_function

import mceq_config as config
from MCEq.core import MCEqRun
import crflux.models as pm
import numpy as np

import pytest
import sys

if sys.platform.startswith("win") and sys.maxsize <= 2**32:
    pytest.skip("Skip model test on 32-bit Windows.", allow_module_level=True)

def format_8_digits(a_list):
    return ["%.8e" % member for member in a_list]

config.debug_level = 1
config.kernel_config = 'numpy'
config.cuda_gpu_id = 0
if config.has_mkl:
    config.set_mkl_threads(2)


mceq = MCEqRun(
    interaction_model='SIBYLL23C',
    theta_deg=0.,
    primary_model=(pm.HillasGaisser2012, 'H3a'))


def test_config_and_file_download():
    import mceq_config as config
    import os
    # Import of config triggers data download
    assert config.mceq_db_fname in os.listdir(config.data_dir)



def test_some_angles():

    nmu = []
    for theta in [0., 30., 60., 90]:
        mceq.set_theta_deg(theta)
        mceq.solve()
        nmu.append(
            np.sum(
                mceq.get_solution('mu+', mag=0, integrate=True) +
                mceq.get_solution('mu-', mag=0, integrate=True)))
    assert format_8_digits(nmu), ['5.62504370e-03', '4.20479234e-03', '1.36630552e-03', '8.20255259e-06']


def test_switch_interaction_models():
    mlist = [
        'DPMJETIII191',
        'DPMJETIII306',
        'QGSJET01C',
        'QGSJETII03',
        'QGSJETII04',
        'SIBYLL21',
        'SIBYLL23',
        'SIBYLL23C03',
        'SIBYLL23C',
        'SIBYLL23CPP']
    count_part = []
    for m in mlist:
        mceq.set_interaction_model(m)
        count_part.append(len(mceq._particle_list))
    assert(count_part == [64, 64, 58, 44, 44, 48, 62, 62, 62, 62])


def test_single_primary():
    energies = [1e3, 1e6, 1e9, 5e10]
    nmu, nnumu, nnue = [], [], []
    mceq.set_interaction_model('SIBYLL23C')
    mceq.set_theta_deg(0.)
    for e in energies:
        mceq.set_single_primary_particle(E=e, pdg_id=2212)
        mceq.solve()
        nmu.append(
            np.sum(
                mceq.get_solution('mu+', mag=0, integrate=True) +
                mceq.get_solution('mu-', mag=0, integrate=True)))
        nnumu.append(
            np.sum(
                mceq.get_solution('numu', mag=0, integrate=True) +
                mceq.get_solution('antinumu', mag=0, integrate=True)))
        nnue.append(
            np.sum(
                mceq.get_solution('nue', mag=0, integrate=True) +
                mceq.get_solution('antinue', mag=0, integrate=True)))
    assert format_8_digits(nmu) == ['2.03134720e+01', '1.20365838e+04', '7.09254150e+06', '2.63982133e+08']
    assert format_8_digits(nnumu) == ['6.80367347e+01', '2.53158948e+04', '1.20884925e+07', '4.14935240e+08']
    assert format_8_digits(nnue) == ['2.36908717e+01', '6.91213253e+03', '2.87396649e+06', '9.27683105e+07']




from ctypes import (cdll, Structure, c_int, c_double, POINTER)
import os
import sysconfig

base = os.path.dirname(os.path.abspath(__file__))
suffix =  sysconfig.get_config_var('EXT_SUFFIX')
# Some Python 2.7 versions don't define EXT_SUFFIX
if suffix is None and 'SO' in sysconfig.get_config_vars():
    suffix =  sysconfig.get_config_var('SO')

assert suffix is not None, 'Shared lib suffix was not identified.'

for fn in os.listdir(base):
    if 'libcorsikaatm' in fn and fn.endswith(suffix):
        corsika_acc = cdll.LoadLibrary(os.path.join(base, fn))
        break

for func in [
        corsika_acc.corsika_get_density, corsika_acc.planar_rho_inv,
        corsika_acc.corsika_get_m_overburden
]:
    func.restype = c_double


def corsika_get_density(h_cm, a, b, c, t, hl):
    """Wrap arguments for ctypes function"""
    return corsika_acc.corsika_get_density(
        c_double(h_cm), a.ctypes, b.ctypes, c.ctypes, t.ctypes, hl.ctypes)


def planar_rho_inv(X, cos_theta, a, b, c, t, hl):
    """Wrap arguments for ctypes function"""
    return corsika_acc.planar_rho_inv(
        c_double(X), c_double(cos_theta), a.ctypes, b.ctypes, c.ctypes,
        t.ctypes, hl.ctypes)


def corsika_get_m_overburden(h_cm, a, b, c, t, hl):
    """Wrap arguments for ctypes function"""
    return corsika_acc.corsika_get_m_overburden(
        c_double(h_cm), a.ctypes, b.ctypes, c.ctypes, t.ctypes, hl.ctypes)




import numpy as np
import mceq_config as config
from MCEq.misc import info


def solv_numpy(nsteps, dX, rho_inv, int_m, dec_m, phi, grid_idcs):
    """:mod:`numpy` implementation of forward-euler integration.

    Args:
      nsteps (int): number of integration steps
      dX (numpy.array[nsteps]): vector of step-sizes :math:`\\Delta X_i` in g/cm**2
      rho_inv (numpy.array[nsteps]): vector of density values :math:`\\frac{1}{\\rho(X_i)}`
      int_m (numpy.array): interaction matrix :eq:`int_matrix` in dense or sparse representation
      dec_m (numpy.array): decay  matrix :eq:`dec_matrix` in dense or sparse representation
      phi (numpy.array): initial state vector :math:`\\Phi(X_0)`
    Returns:
      numpy.array: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    grid_sol = []
    grid_step = 0

    imc = int_m
    dmc = dec_m
    dxc = dX
    ric = rho_inv
    phc = phi

    dXaccum = 0.

    from time import time
    start = time()

    for step in range(nsteps):
        phc += (imc.dot(phc) + dmc.dot(ric[step] * phc)) * dxc[step]

        dXaccum += dxc[step]

        if (grid_idcs and grid_step < len(grid_idcs)
                and grid_idcs[grid_step] == step):
            grid_sol.append(np.copy(phc))
            grid_step += 1

    info(
        2, "Performance: {0:6.2f}ms/iteration".format(1e3 * (time() - start) /
                                                      float(nsteps)))

    return phc, np.array(grid_sol)


class CUDASparseContext(object):
    """This class handles the transfer between CPU and GPU memory, and the calling
    of GPU kernels. Initialized by :class:`MCEq.core.MCEqRun` and used by
    :func:`solv_CUDA_sparse`.
    """
    def __init__(self, int_m, dec_m, device_id=0):

        if config.cuda_fp_precision == 32:
            self.fl_pr = np.float32
        elif config.cuda_fp_precision == 64:
            self.fl_pr = np.float64
        else:
            raise Exception(
                "CUDASparseContext(): Unknown precision specified.")
        # Setup GPU stuff and upload data to it
        try:
            import cupy as cp
            import cupyx.scipy as cpx
            self.cp = cp
            self.cpx = cpx
            self.cubl = cp.cuda.cublas
        except ImportError:
            raise Exception("solv_CUDA_sparse(): Numbapro CUDA libaries not " +
                            "installed.\nCan not use GPU.")

        cp.cuda.Device(config.cuda_gpu_id).use()
        self.cubl_handle = self.cubl.create()
        self.set_matrices(int_m, dec_m)

    def set_matrices(self, int_m, dec_m):
        """Upload sparce matrices to GPU memory"""
        self.cu_int_m = self.cpx.sparse.csr_matrix(int_m, dtype=self.fl_pr)
        self.cu_dec_m = self.cpx.sparse.csr_matrix(dec_m, dtype=self.fl_pr)
        self.cu_delta_phi = self.cp.zeros(self.cu_int_m.shape[0],
                                          dtype=self.fl_pr)

    def alloc_grid_sol(self, dim, nsols):
        """Allocates memory for intermediate if grid solution requested."""
        self.curr_sol_idx = 0
        self.grid_sol = self.cp.zeros((nsols, dim))

    def dump_sol(self):
        """Saves current solution to a new index in grid solution memory."""
        self.cp.copyto(self.grid_sol[self.curr_sol_idx, :], self.cu_curr_phi)
        self.curr_sol_idx += 1
        # self.grid_sol[self.curr_sol, :] = self.cu_curr_phi

    def get_gridsol(self):
        """Downloads grid solution to main memory."""
        return self.cp.asnumpy(self.grid_sol)

    def set_phi(self, phi):
        """Uploads initial condition to GPU memory."""
        self.cu_curr_phi = self.cp.asarray(phi, dtype=self.fl_pr)

    def get_phi(self):
        """Downloads current solution from GPU memory."""
        return self.cp.asnumpy(self.cu_curr_phi)

    def solve_step(self, rho_inv, dX):
        """Makes one solver step on GPU using cuSparse (BLAS)"""

        self.cp.cusparse.csrmv(a=self.cu_int_m,
                               x=self.cu_curr_phi,
                               y=self.cu_delta_phi,
                               alpha=1.,
                               beta=0.)
        self.cp.cusparse.csrmv(a=self.cu_dec_m,
                               x=self.cu_curr_phi,
                               y=self.cu_delta_phi,
                               alpha=rho_inv,
                               beta=1.)
        self.cubl.saxpy(self.cubl_handle, self.cu_delta_phi.shape[0], dX,
                        self.cu_delta_phi.data.ptr, 1,
                        self.cu_curr_phi.data.ptr, 1)


def solv_CUDA_sparse(nsteps, dX, rho_inv, context, phi, grid_idcs):
    """`NVIDIA CUDA cuSPARSE <https://developer.nvidia.com/cusparse>`_ implementation
    of forward-euler integration.

    Function requires a working :mod:`accelerate` installation.

    Args:
      nsteps (int): number of integration steps
      dX (numpy.array[nsteps]): vector of step-sizes :math:`\\Delta X_i` in g/cm**2
      rho_inv (numpy.array[nsteps]): vector of density values :math:`\\frac{1}{\\rho(X_i)}`
      int_m (numpy.array): interaction matrix :eq:`int_matrix` in dense or sparse representation
      dec_m (numpy.array): decay  matrix :eq:`dec_matrix` in dense or sparse representation
      phi (numpy.array): initial state vector :math:`\\Phi(X_0)`
      mu_loss_handler (object): object of type :class:`SemiLagrangianEnergyLosses`
    Returns:
      numpy.array: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    c = context
    c.set_phi(phi)

    grid_step = 0

    from time import time
    start = time()
    if len(grid_idcs) > 0:
        c.alloc_grid_sol(phi.shape[0], len(grid_idcs))

    for step in range(nsteps):
        c.solve_step(rho_inv[step], dX[step])

        if (grid_idcs and grid_step < len(grid_idcs)
                and grid_idcs[grid_step] == step):
            c.dump_sol()
            grid_step += 1

    info(
        2, "Performance: {0:6.2f}ms/iteration".format(1e3 * (time() - start) /
                                                      float(nsteps)))

    return c.get_phi(), c.get_gridsol() if len(grid_idcs) > 0 else []


def solv_MKL_sparse(nsteps, dX, rho_inv, int_m, dec_m, phi, grid_idcs):
    # mu_loss_handler):
    """`Intel MKL sparse BLAS
    <https://software.intel.com/en-us/articles/intel-mkl-sparse-blas-overview?language=en>`_
    implementation of forward-euler integration.

    Function requires that the path to the MKL runtime library ``libmkl_rt.[so/dylib]``
    defined in the config file.

    Args:
      nsteps (int): number of integration steps
      dX (numpy.array[nsteps]): vector of step-sizes :math:`\\Delta X_i` in g/cm**2
      rho_inv (numpy.array[nsteps]): vector of density values :math:`\\frac{1}{\\rho(X_i)}`
      int_m (numpy.array): interaction matrix :eq:`int_matrix` in dense or sparse representation
      dec_m (numpy.array): decay  matrix :eq:`dec_matrix` in dense or sparse representation
      phi (numpy.array): initial state vector :math:`\\Phi(X_0)`
      grid_idcs (list): indices at which longitudinal solutions have to be saved.

    Returns:
      numpy.array: state vector :math:`\\Phi(X_{nsteps})` after integration
    """

    from ctypes import c_int, c_char, POINTER, byref
    from mceq_config import mkl

    gemv = None
    axpy = None
    np_fl = None
    from ctypes import c_double as fl_pr
    # sparse CSR-matrix x dense vector
    gemv = mkl.mkl_dcsrmv
    # dense vector + dense vector
    axpy = mkl.cblas_daxpy
    np_fl = np.float64

    # Prepare CTYPES pointers for MKL sparse CSR BLAS
    int_m_data = int_m.data.ctypes.data_as(POINTER(fl_pr))
    int_m_ci = int_m.indices.ctypes.data_as(POINTER(c_int))
    int_m_pb = int_m.indptr[:-1].ctypes.data_as(POINTER(c_int))
    int_m_pe = int_m.indptr[1:].ctypes.data_as(POINTER(c_int))

    dec_m_data = dec_m.data.ctypes.data_as(POINTER(fl_pr))
    dec_m_ci = dec_m.indices.ctypes.data_as(POINTER(c_int))
    dec_m_pb = dec_m.indptr[:-1].ctypes.data_as(POINTER(c_int))
    dec_m_pe = dec_m.indptr[1:].ctypes.data_as(POINTER(c_int))

    npphi = np.copy(phi).astype(np_fl)
    phi = npphi.ctypes.data_as(POINTER(fl_pr))
    npdelta_phi = np.zeros_like(npphi)
    delta_phi = npdelta_phi.ctypes.data_as(POINTER(fl_pr))

    trans = c_char(b'n')
    npmatd = np.chararray(6)
    npmatd[0] = b'G'
    npmatd[3] = b'C'
    matdsc = npmatd.ctypes.data_as(POINTER(c_char))
    m = c_int(int_m.shape[0])
    cdzero = fl_pr(0.)
    cdone = fl_pr(1.)
    cione = c_int(1)

    grid_step = 0
    grid_sol = []

    from time import time
    start = time()

    for step in range(nsteps):
        # delta_phi = int_m.dot(phi)
        gemv(byref(trans), byref(m), byref(m), byref(cdone),
             matdsc, int_m_data, int_m_ci, int_m_pb, int_m_pe, phi,
             byref(cdzero), delta_phi)
        # delta_phi = rho_inv * dec_m.dot(phi) + delta_phi
        gemv(byref(trans), byref(m), byref(m), byref(fl_pr(rho_inv[step])),
             matdsc, dec_m_data, dec_m_ci, dec_m_pb, dec_m_pe, phi,
             byref(cdone), delta_phi)
        # phi = delta_phi * dX + phi
        axpy(m, fl_pr(dX[step]), delta_phi, cione, phi, cione)

        if (grid_idcs and grid_step < len(grid_idcs)
                and grid_idcs[grid_step] == step):
            grid_sol.append(np.copy(npphi))
            grid_step += 1

    info(
        2, "Performance: {0:6.2f}ms/iteration".format(1e3 * (time() - start) /
                                                      float(nsteps)))

    return npphi, np.asarray(grid_sol)


# # TODO: Debug this and transition to BDF
# def _odepack(dXstep=.1,
#                 initial_depth=0.0,
#                 int_grid=None,
#                 grid_var='X',
#                 *args,
#                 **kwargs):
#     """Solves the transport equations with solvers from ODEPACK.

#     Args:
#         dXstep (float): external step size (adaptive sovlers make more steps internally)
#         initial_depth (float): starting depth in g/cm**2
#         int_grid (list): list of depths at which results are recorded
#         grid_var (str): Can be depth `X` or something else (currently only `X` supported)

#     """
#     from scipy.integrate import ode
#     ri = self.density_model.r_X2rho

#     if config.enable_muon_energy_loss:
#         raise NotImplementedError(
#             'Energy loss not imlemented for this solver.')

#     # Functional to solve
#     def dPhi_dX(X, phi, *args):
#         return self.int_m.dot(phi) + self.dec_m.dot(ri(X) * phi)

#     # Jacobian doesn't work with sparse matrices, and any precision
#     # or speed advantage disappear if used with dense algebra
#     def jac(X, phi, *args):
#         # print 'jac', X, phi
#         return (self.int_m + self.dec_m * ri(X)).todense()

#     # Initial condition
#     phi0 = np.copy(self.phi0)

#     # Initialize variables
#     grid_sol = []

#     # Setup solver
#     r = ode(dPhi_dX).set_integrator(
#         with_jacobian=False, **config.ode_params)

#     if int_grid is not None:
#         initial_depth = int_grid[0]
#         int_grid = int_grid[1:]
#         max_X = int_grid[-1]
#         grid_sol.append(phi0)

#     else:
#         max_X = self.density_model.max_X

#     info(
#         1,
#         'your X-grid is shorter then the material',
#         condition=max_X < self.density_model.max_X)
#     info(
#         1,
#         'your X-grid exceeds the dimentsions of the material',
#         condition=max_X > self.density_model.max_X)

#     # Initial value
#     r.set_initial_value(phi0, initial_depth)

#     info(
#         2, 'initial depth: {0:3.2e}, maximal depth {1:}'.format(
#             initial_depth, max_X))

#     start = time()
#     if int_grid is None:
#         i = 0
#         while r.successful() and (r.t + dXstep) < max_X - 1:
#             info(5, "Solving at depth X =", r.t, condition=(i % 5000) == 0)
#             r.integrate(r.t + dXstep)
#             i += 1
#         if r.t < max_X:
#             r.integrate(max_X)
#         # Do last step to make sure the rational number max_X is reached
#         r.integrate(max_X)
#     else:
#         for i, Xi in enumerate(int_grid):
#             info(5, 'integrating at X =', Xi, condition=i % 10 == 0)

#             while r.successful() and (r.t + dXstep) < Xi:
#                 r.integrate(r.t + dXstep)

#             # Make sure the integrator arrives at requested step
#             r.integrate(Xi)
#             # Store the solution on grid
#             grid_sol.append(r.y)

#     info(2,
#             'time elapsed during integration: {1} sec'.format(time() - start))

#     self.solution = r.y
#     self.grid_sol = grid_sol
